
{-
mamot: Memories Are Made Of This
Copyright (C) 2004  Ian Lynagh <igloo@earth.li>
-}

module Main (main) where

import Graphics.UI.WXCore hiding (Time)
import Graphics.UI.WX
import Char (isDigit)
import List (insert)
import Maybe (fromMaybe)
import Monad (mplus)
import System (getArgs, exitWith, ExitCode(..))

data Memory = Memory {mid :: ID,
                      begin :: Date,
                      end :: Date,
                      begin_fuzz :: Days,
                      end_fuzz :: Days,
                      level :: Level,
                      desc :: String}
    deriving (Read, Show)

instance Eq Memory where
    m1 == m2 = mid m1 == mid m2

instance Ord Memory where
    m1 <= m2 = mid m1 <= mid m2

data Info = Info { max_level :: Int,
                   next_mid :: Int,
                   last_date :: Date,
                   full_height :: Int,
                   height :: Int,
                   filename :: Maybe FilePath,
                   export_filename :: Maybe FilePath,
                   im_width :: Int,
                   last_added :: Maybe Memory,
                   selected :: Maybe Memory
                 }

instance Eq Date where
    d1 == d2 = to_time d1 == to_time d2

instance Ord Date where
    d1 <= d2 = to_time d1 <= to_time d2

data Date = Date { _day :: Int, _month :: Int, year :: Int }

instance Read Date where
    readsPrec _ s = case break (== '/') s of
                     (dd, '/':s') ->
                      case break (== '/') s' of
                       (mm, '/':s'') ->
                        case span isDigit s'' of
                         (yyyy, s''') ->
                          case (read' dd, read' mm, read' yyyy) of
                           (Just dd', Just mm', Just yyyy') ->
                            [(Date dd' mm' yyyy', s''')]
                           _ -> []
                       _ -> []
                     _ -> []

instance Show Date where
    showsPrec _ (Date dd mm yyyy) = showsPrec 0 dd . showChar '/'
                                  . showsPrec 0 mm . showChar '/'
                                  . showsPrec 0 yyyy

type Memories = ([Memory], Info)

type Time = Rational
type Days = Int
type Level = Int
type ID = Int

curviness :: Double
curviness = 3
text_hinset, text_hpadding, text_vinset, hgap, vgap :: Int
text_hinset = 3
text_hpadding = 2 * text_hinset
text_vinset = 1
hgap = 3
vgap = 3

initial_width :: Int
initial_width = 400

main :: IO ()
main = start disp

bug :: String -> a
bug s = error $ "mamot: panic: this can't happen!\n" ++ s ++ "\n"

disp :: IO ()
disp = do f <- frame [text := "Memories Are Made Of This"]
          s <- scrolledWindow f [virtualSize := sz initial_width 0,
                                 scrollRate := sz 1 1
                                 ]
          let redraw = repaint s
          ext <- withClientDC s (\dc -> getTextExtent dc "Fy")
          let h = sizeH ext + text_vinset
              i = Info {
                        next_mid = 0,
                        max_level = 0,
                        last_date = read "0/0/0", -- Has to be valid for show,
                                                  -- but is never really used
                        full_height = (h+vgap),
                        height = h,
                        filename = Nothing,
                        export_filename = Nothing,
                        selected = Nothing,
                        last_added = Nothing,
                        im_width = initial_width
                       }
          v <- varCreate ([], i)
          m_file <- menuPane [text := "&File"]
          menuItem m_file [text := "&New\tCtrl+N",
                           on command := new v redraw]
          menuItem m_file [text := "&Load...\tCtrl+L",
                           on command := load f v redraw]
          menuItem m_file [text := "&Reload\tCtrl+R",
                           on command := reload f v redraw]
          menuLine m_file
          menuItem m_file [text := "&Export",
                           on command := export f v]
          menuItem m_file [text := "&Export to...",
                           on command := export_to f v]
          menuItem m_file [text := "&Save\tCtrl+S",
                           on command := save f v]
          menuItem m_file [text := "Save &As...",
                           on command := save_as f v]
          menuLine m_file
          menuItem m_file [text := "&Quit\tCtrl+Q",
                           on command := close f]
          m_view <- menuPane [text := "&View"]
          menuItem m_view [text := "Zoom &In\tCtrl++",
                           on command := zoom_in v s redraw]
          menuItem m_view [text := "Zoom &Out\tCtrl+-",
                           on command := zoom_out v s redraw]
          m_memory <- menuPane [text := "&Memory"]
          menuItem m_memory [text := "&Add...\tCtrl+A",
                             on command := add_memory f v redraw]
          menuItem m_memory [text := "&Edit...\tCtrl+E",
                             on command := edit_memory f v redraw]
          menuItem m_memory [text := "&Delete\tCtrl+D",
                             on command := delete_memory f v redraw]
          set s [on (charKey 'q') := close f,
                 on click := clicked redraw v,
                 on paint := paint_events (Just s) v]
          set f [layout := fill (widget s),
                 menuBar := [m_file, m_view, m_memory]]
          handle_args f v redraw

handle_args :: Window a -> Var Memories -> IO () -> IO ()
handle_args w v redraw
 = do args <- getArgs
      case args of
          [f] -> do (ms, i) <- varGet v
                    varSet v (ms, i {filename = Just f})
                    reload w v redraw
          [] -> return ()
          _ -> do errorDialog w "Error" "Bad arguments"
                  exitWith (ExitFailure 1)

paint_events :: Maybe (ScrolledWindow a) -> Var Memories -> DC a
             -> Rect -> IO ()
paint_events m_sw v dc _
 = do msi@(ms, i) <- varGet v
      let to_coord = make_to_coord msi
          h = image_height i
      case m_sw of
          Just sw -> set sw [virtualSize := size (im_width i) h]
          Nothing -> return ()
      set dc [bgcolor := white]
      dcClear dc
      set dc [textColor := black]
      mapM_ (draw_event to_coord dc i) ms
      case ms of
          (m:_) -> print_scale to_coord dc (begin_fuzzy_date m) i
          _ -> return ()

make_to_coord :: Memories -> (Time -> Int)
make_to_coord (ms, i)
 = let min_t = begin_fuzzy_time (head ms)
       max_t = to_time (last_date i)
       width = fromIntegral (im_width i - 2 * hgap)
   in \t -> hgap + round (width * (t - min_t) / (max_t - min_t))

draw_event :: (Time -> Int) -> DC a -> Info -> Memory -> IO ()
draw_event to_coord dc i m
 = do set dc [brushColor := white, brushKind := BrushSolid]
      roundedRect dc (rectBetween tl' br') curviness []
      let colour = if Just m == selected i then cyan else lightgrey
      set dc [brushColor := colour]
      roundedRect dc (rectBetween tl br) curviness []
      ext <- getTextExtent dc (desc m)
      when (sizeW ext + text_hpadding <= rc - lc)
         $ drawText dc (desc m) text_tl []
    where tl' = pt (to_coord (begin_fuzzy_time m)) (bottom - height i)
          br' = pt (to_coord (end_fuzzy_time m)) bottom
          text_tl = pt (to_coord (begin_time m) + text_hinset)
                       (bottom - height i + text_vinset)
          tl = pt lc (bottom - height i)
          br = pt rc bottom
          lc = to_coord (begin_time m)
          rc = to_coord (end_time m)
          bottom = (1 + max_level i - level m) * full_height i

print_scale :: (Time -> Int) -> DC a -> Date -> Info -> IO ()
print_scale to_coord dc start_date i
 = do let years = [year start_date .. year (last_date i) + 1]
          text_y_coord = (2 + max_level i) * full_height i
          bot_y_coord = (2 + max_level i) * full_height i
          top_y_coord = bot_y_coord - height i
          mid_y_coord = bot_y_coord - (height i `div` 2)
          draw_year y = do let s = show y
                           ext <- getTextExtent dc s
                           let x_coord = to_coord (to_time (Date 0 1 y))
                               text_x_coord = x_coord - (sizeW ext `div` 2)
                               far_right =    x_coord + (sizeW ext `div` 2)
                               text_pt = pt text_x_coord text_y_coord
                           when (text_x_coord >= 0 && far_right <= im_width i)
                              $ do drawText dc s text_pt []
                           when (x_coord >= 0 && x_coord <= im_width i)
                              $ line dc (pt x_coord top_y_coord)
                                        (pt x_coord mid_y_coord) []
      set dc [penColor := black]
      mapM_ draw_year years
      set dc [penColor := red]
      line dc (pt 0 top_y_coord) (pt (im_width i) top_y_coord) []

image_height :: Info -> Int
image_height i = (3 + max_level i) * full_height i

insert_memory :: Memory -> Memories -> Memories
insert_memory m (ms, i) = insert_memory' (m {mid = next_mid i})
                                         (ms, i { next_mid = next_mid i + 1,
                                                  last_date = new_last_date})
    where this_end_fuzzy_date = end_fuzzy_date m
          cur_last_date = last_date i
          new_last_date = if to_time this_end_fuzzy_date > to_time cur_last_date
                          then this_end_fuzzy_date
                          else cur_last_date

insert_memory' :: Memory -> Memories -> Memories
insert_memory' m (ms, i) = assign_levels (ins m ms, i)
    where ins x [] = [x]
          ins x ys@(y:ys') = if begin_fuzzy_time x <= begin_fuzzy_time y
                             then x:ys
                             else y:ins x ys'

assign_levels :: Memories -> Memories
assign_levels (ms, i) = let ms' = assign_levels' ([], 0) ms
                        in (ms', i {max_level = maximum (map level ms')})

assign_levels' :: ([(Time, Level)], Level) -> [Memory] -> [Memory]
assign_levels' _ [] = []
assign_levels' st (m:ms)
 = case get_level st (begin_fuzzy_time m) of
       ((tls, nl), l) ->
           m {level = l}:assign_levels' (insert (end_fuzzy_time m, l) tls, nl) ms

get_level :: ([(Time, Level)], Level) -> Time
          -> (([(Time, Level)], Level), Level)
get_level ([], l) _ = (([], l+1), l)
get_level ((t, l):tls, next_l) this_t
 | t <= this_t = ((tls, next_l), l)
 | otherwise   = case get_level (tls, next_l) this_t of
                     ((tls', next_l'), l') -> (((t, l):tls', next_l'), l')

begin_fuzzy_time :: Memory -> Time
begin_fuzzy_time m = to_time (begin_fuzzy_date m)

begin_fuzzy_date :: Memory -> Date
begin_fuzzy_date m = case begin m of
                         Date dd mm yyyy ->
                             Date (dd - begin_fuzz m) mm yyyy

begin_time :: Memory -> Time
begin_time m = to_time (begin m)

end_time :: Memory -> Time
end_time m = to_time (end m)

end_fuzzy_time :: Memory -> Time
end_fuzzy_time m = to_time (end_fuzzy_date m)

end_fuzzy_date :: Memory -> Date
end_fuzzy_date m = case end m of
                       Date dd mm yyyy -> Date (dd + end_fuzz m) mm yyyy

to_time :: Date -> Time
to_time d@(Date _ _ yyyy) = fromIntegral yyyy
                          +     fromIntegral (days_till d)
                              / fromIntegral (days_in_year yyyy)

days_in_year :: Int -> Days
days_in_year n = if is_leap_year n then 366 else 365

days_till :: Date -> Days
days_till (Date dd mm yyyy) = dd + sum (take (mm - 1) mons)
    where mons = [31, feb, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
          feb = if is_leap_year yyyy then 29 else 28

is_leap_year :: Int -> Bool
is_leap_year n = (n `mod` 400 == 0) || ((n `mod` 100 /= 0) && (n `mod` 4 == 0))

read' :: Read a => String -> Maybe a
read' x = case reads x of
              [(y, "")] -> Just y
              [(y, "\n")] -> Just y
              _ -> Nothing

----- Selecting a memory

clicked :: IO () -> Var Memories -> Point -> IO ()
clicked redraw v p
 = do msi@(ms, i) <- varGet v
      let l = max_level i - (pointY p `div` full_height i)
          ms' = filter ((l ==) . level) ms
          sel = get_memory_at (make_to_coord msi) (pointX p) ms'
      varSet v (ms, i { selected = sel })
      redraw

get_memory_at :: (Time -> Int) -> Int -> [Memory] -> Maybe Memory
get_memory_at _ _ [] = Nothing
get_memory_at to_coord x (m:ms)
 | to_coord (begin_fuzzy_time m) <= x && to_coord (end_fuzzy_time m) >= x
        = Just m
 | otherwise = get_memory_at to_coord x ms

----- Add a memory

add_memory :: Window a -> Var Memories -> IO () -> IO ()
add_memory w v redraw
 = do (ms, i) <- varGet v
      m_m <- memory_dialog w (selected i `mplus` last_added i) "Add a memory"
      case m_m of
          Just m -> do varSet v (insert_memory m (ms, i {last_added = m_m}))
                       redraw
          Nothing -> return ()

edit_memory :: Window a -> Var Memories -> IO () -> IO ()
edit_memory w v redraw
 = do (ms, i) <- varGet v
      case selected i of
          Just m -> do
              m_m <- memory_dialog w (Just m) "Edit memory"
              case m_m of
                  Just m' -> do let ms' = filter (/= m) ms
                                    (ms'', i') = insert_memory' (m' {mid = mid m})
                                                                 (ms', i)
                                    new_last_date = maximum (map end_fuzzy_date ms'')
                                varSet v (ms'', i' {last_date = new_last_date})
                                redraw
                  Nothing -> return ()
          Nothing -> err "No memory selected"
    where err s = errorDialog w "Error" s

memory_dialog :: Window a -> Maybe Memory -> String -> IO (Maybe Memory)
memory_dialog w m title
 = do d <- dialog w [text := title]
      p <- panel d []
      let get_str  f = fromMaybe "" (fmap f m)
          get_show f = get_str (show . f)
          select_all c = textCtrlSetSelection c (-1) (-1)
      t_desc       <- textEntry p AlignLeft [text := get_str  desc]
      t_begin_fuzz <- textEntry p AlignLeft [text := get_show begin_fuzz]
      t_begin      <- textEntry p AlignLeft [text := get_show begin]
      t_end        <- textEntry p AlignLeft [text := get_show end]
      t_end_fuzz   <- textEntry p AlignLeft [text := get_show end_fuzz]
      mapM_ select_all [t_begin_fuzz, t_begin, t_end, t_end_fuzz, t_desc]
      ok     <- button p [text := "Ok"]
      cancel <- button p [text := "Cancel"]

      let l_desc = row 5 [label "Description:", hfill $ widget t_desc]
          l_time = boxed "When"
               $ grid 5 5 [
           [label "Begin fuzz:", hfill $ widget t_begin_fuzz, label "days"],
           [label "Begin:", hfill $ widget t_begin, label "(dd/mm/yyyy)"],
           [label "End:",   hfill $ widget t_end, label "(dd/mm/yyyy)"],
           [label "End fuzz:",       hfill $ widget t_end_fuzz, label "days"]]
          l_buttons = floatBottomRight $ row 5 [vfill (widget ok),
                                                vfill (widget cancel)]

      buttonSetDefault ok
      set d [layout := container p $ margin 10
                                   $ column 10 [l_desc, l_time, l_buttons]]
      focusOn t_desc
      let err s = errorDialog d "Error" s
      showModal d $ \stop ->
          do focusOn t_desc
             set ok [on command := do
                new_desc <- get t_desc text
                new_begin_fuzz <- get t_begin_fuzz text
                new_begin <- get t_begin text
                new_end <- get t_end text
                new_end_fuzz <- get t_end_fuzz text
                case read' ('0':new_begin_fuzz) of
                 Just new_begin_fuzz' ->
                  case read' new_begin of
                   Just new_begin' ->
                    case read' new_end of
                     Just new_end' ->
                      case read' ('0':new_end_fuzz) of
                       Just new_end_fuzz' ->
                        stop (Just (Memory {
                                        mid = bug "Main.memory_dialog: mid",
                                        begin = new_begin',
                                        end = new_end',
                                        begin_fuzz = new_begin_fuzz',
                                        end_fuzz = new_end_fuzz',
                                        level = bug "Main.memory_dialog: level",
                                        desc = new_desc }))
                       Nothing -> err "Bad end fuzz"
                     Nothing -> err "Bad end date"
                   Nothing -> err "Bad begin date"
                 Nothing -> err "Bad begin fuzz"
              ]
             set cancel [on command := stop Nothing]

----- Delete selected memory

delete_memory :: Window a -> Var Memories -> IO () -> IO ()
delete_memory w v redraw
 = do (ms, i) <- varGet v
      case selected i of
          Nothing -> err "No memory selected"
          Just m -> do varSet v (assign_levels (remove_memory m ms,
                                                i {next_mid = next_mid i - 1,
                                                   selected = Nothing}))
                       redraw
    where err s = errorDialog w "Error" s

remove_memory :: Memory -> [Memory] -> [Memory]
remove_memory _ [] = []
remove_memory i (m:ms)
 | m == i = rest
 | m > i = m {mid = mid m - 1}:rest
 | otherwise = m:rest
    where rest = remove_memory i ms

----- New, Loading and Saving

save :: Window a -> Var Memories -> IO ()
save w v = do (ms, i) <- varGet v
              case filename i of
                  Nothing -> save_as w v
                  Just f ->
                      do let save_me = (ms, max_level i, next_mid i, last_date i)
                         catch (writeFile f (show save_me))
                               (\_ -> err ("Failed to save to " ++ f))
    where err s = errorDialog w "Error" s

save_as :: Window a -> Var Memories -> IO ()
save_as w v = do (ms, i) <- varGet v
                 let fn = fromMaybe "" (filename i)
                 m_fn' <- fileSaveDialog w False True "Save as..." "" fn
                 case m_fn' of
                     Just fn' -> do varSet v (ms, i {filename = Just fn'})
                                    save w v
                     Nothing -> return ()

new :: Var Memories -> IO () -> IO ()
new v redraw
 = do (_, i) <- varGet v
      varSet v ([], i {max_level = 0,
                       next_mid = 0,
                       last_date = read "0/0/0",
                       selected = Nothing})
      redraw

reload :: Window a -> Var Memories -> IO () -> IO ()
reload w v redraw
 = do (_, i) <- varGet v
      case filename i of
          Nothing -> load w v redraw
          Just f -> catch (do x <- readFile f
                              case read' x of
                                  Just (ms, l, nm, d) ->
                                      do let i' = i {max_level = l,
                                                     next_mid = nm,
                                                     selected = Nothing,
                                                     last_date = d}
                                         varSet v (ms, i')
                                         redraw
                                  Nothing -> err "Bad file contents")
                          (\_ -> err ("Failed to read " ++ f))
    where err s = errorDialog w "Error" s

load :: Window a -> Var Memories -> IO () -> IO ()
load w v redraw
 = do (ms, i) <- varGet v
      let fn = fromMaybe "" (filename i)
      m_fn' <- fileOpenDialog w False True "Load..." [] "" fn
      case m_fn' of
          Just fn' -> do varSet v (ms, i {filename = Just fn'})
                         reload w v redraw
          Nothing -> return ()

export :: Window a -> Var Memories -> IO ()
export w v = do msi@(ms, i) <- varGet v
                let h = image_height i
                case export_filename i of
                    Nothing -> export_to w v
                    Just f ->
                        do varSet v (ms, i { selected = Nothing })
                           catch (do_export f (size (im_width i) h))
                                 (\_ -> err ("Failed to export to " ++ f))
                           varSet v msi
    where err s = errorDialog w "Error" s
          do_export :: FilePath -> Size -> IO ()
          do_export f s
           = do dc <- memoryDCCreate
                i <- imageCreateSized s
                b <- imageConvertToBitmap i
                memoryDCSelectObject dc b
                paint_events Nothing v dc (bug "export: rect")
                r <- bitmapSaveFile b f wxBITMAP_TYPE_PNG nullPalette
                when (r /= 1{-TRUE-}) (err "Failed to write output")
                bitmapDelete b
                memoryDCDelete dc


export_to :: Window a -> Var Memories -> IO ()
export_to w v
 = do (ms, i) <- varGet v
      let fn = fromMaybe "" (export_filename i)
      m_fn' <- fileSaveDialog w False True "Export to..." "" fn
      case m_fn' of
          Just fn' -> do varSet v (ms, i {export_filename = Just fn'})
                         export w v
          Nothing -> return ()

----- Zooming

zoom_in :: Var Memories -> ScrolledWindow a -> IO () -> IO ()
zoom_in = zoom (*2)

zoom_out :: Var Memories -> ScrolledWindow a -> IO () -> IO ()
zoom_out = zoom (\w -> (w `div` 2) `max` (2 * hgap + 1))

zoom :: (Int -> Int) -> Var Memories -> ScrolledWindow a -> IO () -> IO ()
zoom f v sw redraw = do (ms, i) <- varGet v
                        let w = f (im_width i)
                        varSet v (ms, i {im_width = w})
                        set sw [virtualSize := sz w 0]
                        redraw

